import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import '../types/types.dart';

/// Properties identifying the current state of a [VdoPlayerController].
class VdoPlayerValue {
  /// True if a video is currently being loaded into the player.
  final bool isLoading;

  /// Properties related to the currently loaded video.
  final MediaInfo mediaInfo;

  /// Total duration of the video.
  ///
  /// This is [Duration.zero] if the video is not loaded.
  final Duration duration;

  /// The current playback position.
  final Duration position;

  /// Position up to where the video is buffered.
  final Duration bufferedPosition;

  /// True if the video is playing, false if paused.
  final bool isPlaying;

  /// True if the video is buffering.
  final bool isBuffering;

  /// True if the currently loaded video has played to the end.
  final bool isEnded;

  /// The current playback speed.
  final double playbackSpeed;

  /// The currently selected caption language.
  final String captionLanguage;

  /// If not null, this gives details about an error encountered while loading
  /// the video or during playback.
  final VdoError vdoError;

  VdoPlayerValue({
    this.duration = Duration.zero,
    this.position = Duration.zero,
    this.bufferedPosition = Duration.zero,
    this.isLoading = false,
    this.mediaInfo,
    this.isPlaying = false,
    this.isBuffering = false,
    this.isEnded = false,
    this.playbackSpeed = 1.0,
    this.captionLanguage,
    this.vdoError
  });

  /// Returns a new instance with the same values as this current instance,
  /// except for any overrides specified as arguments.
  VdoPlayerValue copyWith({
    bool isLoading,
    MediaInfo mediaInfo,
    Duration duration,
    Duration position,
    Duration bufferedPosition,
    bool isPlaying,
    bool isBuffering,
    bool isEnded,
    double playbackSpeed,
    String captionLanguage,
    VdoError vdoError
  }) {
    return VdoPlayerValue(
      duration: duration ?? this.duration,
      position: position ?? this.position,
      bufferedPosition: bufferedPosition ?? this.bufferedPosition,
      isLoading: isLoading ?? this.isLoading,
      mediaInfo: mediaInfo ?? this.mediaInfo,
      isPlaying: isPlaying ?? this.isPlaying,
      isBuffering: isBuffering ?? this.isBuffering,
      isEnded: isEnded ?? this.isEnded,
      playbackSpeed: playbackSpeed ?? this.playbackSpeed,
      captionLanguage: captionLanguage ?? this.captionLanguage,
      vdoError: vdoError ?? this.vdoError
    );
  }
}

/// Controls a native video player and notifies any listeners of the player
/// state changes.
///
/// The VdoPlayerView widget is used to display the video and it provides
/// an instance of this class to control the video in the onPlayerCreated
/// callback.
class VdoPlayerController extends ValueNotifier<VdoPlayerValue> with WidgetsBindingObserver {
  final MethodChannel _channel;
  final EventChannel _eventChannel;

  Timer _positionTimer;

  VdoPlayerController(this._channel, this._eventChannel)
      : super(VdoPlayerValue());

  void init() {
    _eventChannel.receiveBroadcastStream().listen(
        _handleNativeEvent,
        onError: _handleNativeErrorEvent
    );
    WidgetsBinding.instance.addObserver(this);
  }

  @override
  void dispose() {
    _stopPositionTimer();
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    super.didChangeAppLifecycleState(state);
    // TODO didChangeAppLifecycleState often called multiple times, fix!
    if (state == AppLifecycleState.paused) {
      // Save playback state and release player resources
      _packUp();
    }
    if (state == AppLifecycleState.resumed) {
      // Restore with playback state saved on native side
      _restore();
    }
  }

  bool get isPlaying => value.isPlaying;
  bool get isBuffering => value.isBuffering;
  bool get isLoading => value.isLoading;
  bool get isEnded => value.isEnded;

  /// Load a new video with specified options to the player as identified by the
  /// embedInfo argument.
  Future<LoadResult> load(EmbedInfo embedInfo) async {
    try {
      // renew VdoPlayerValue, remove old video related parameters
      _stopPositionTimer();
      value = VdoPlayerValue(isLoading: true);
      Map<String, Object> loadArgs = {
        'restore' : false,
        'embedInfo' : embedInfo.toMap()
      };
      Map loadResultMap = await _channel.invokeMethod("load", loadArgs);
      print("load success");
      Map mediaInfoMap = loadResultMap['mediaInfo'];

      MediaInfo mediaInfo = MediaInfo(
          mediaId: mediaInfoMap['mediaId'],
          type: MediaInfoType.streaming, // TODO fix
          title: mediaInfoMap['title'],
          description: mediaInfoMap['description'],
          duration: Duration(milliseconds: mediaInfoMap['duration']));

      value = value.copyWith(
          isLoading: false,
          mediaInfo: mediaInfo,
          duration: mediaInfo.duration,
          isPlaying: loadResultMap['playWhenReady']);

      if (value.isPlaying) {
        _startPositionTimer();
      }

      return LoadResult(mediaInfo: mediaInfo, isPlaying: value.isPlaying);
    } on PlatformException catch (e) { // TODO
      print("load error [" + e.code + ", " + e.message + "]");
      VdoError error = VdoError(int.parse(e.code), e.message, null);
      value = value.copyWith(
          isLoading: false,
          isBuffering: false,
          vdoError: error
      );
      throw e;
    }
  }

  /// Start or resume playing the video.
  Future<void> play() async {
    value = value.copyWith(isPlaying: true);
    _startPositionTimer();
    await _channel.invokeMethod("setPlayWhenReady", isPlaying);
  }

  /// Pause video playback.
  Future<void> pause() async {
    value = value.copyWith(isPlaying: false);
    _stopPositionTimer();
    await _channel.invokeMethod("setPlayWhenReady", isPlaying);
  }

  /// Unload the currently loaded video from the player.
  Future<void> stop() async {
    _stopPositionTimer();
    await _channel.invokeMethod("stop");
  }

  /// Returns the total duration of the video.
  Future<Duration> getDuration() async {
    int durationMs = await _channel.invokeMethod("getDuration");
    return Duration(milliseconds: durationMs);
  }

  /// Returns the current playback position.
  Future<Duration> getPosition() async {
    List positions = await _getPlaybackAndBufferedPositions();
    return positions[0];
  }

  /// Seek to the specified target position.
  Future<void> seek(Duration target) async {
    target = target > value.duration ? value.duration : target;
    await _channel.invokeMethod("seek", target.inMilliseconds);
  }

  /// Set the video playback speed.
  Future<void> setPlaybackSpeed(double speed) async {
    value = value.copyWith(playbackSpeed: speed);
    await _channel.invokeMethod("setPlaybackSpeed", speed);
  }

  /// Set caption language. Pass null to disable captions.
  ///
  /// Returns true if a caption of the specified language was found and applied,
  /// false otherwise.
  Future<bool> setCaptionLanguage(String language) async {
    bool changed = await _channel.invokeMethod("setCaptionLanguage", language);
    if (changed) {
      value = value.copyWith(captionLanguage: language);
      return true;
    } else {
      print('Caption language not changed.');
      return false;
    }
  }

  _startPositionTimer() {
    if (_positionTimer != null) {
      _positionTimer.cancel();
    }
    _positionTimer = Timer.periodic(
        Duration(milliseconds: 500),
            (timer) async {
          List<Duration> newPositions = await _getPlaybackAndBufferedPositions();
          value = value.copyWith(
            position: newPositions[0],
            bufferedPosition: newPositions[1]
          );
        });
  }

  _stopPositionTimer() {
    if (_positionTimer != null) {
      _positionTimer.cancel();
    }
  }

  Future<List<Duration>> _getPlaybackAndBufferedPositions() async {
    Map positions = await _channel.invokeMethod("getPlaybackAndBufferedPositions");
    int positionMs = positions['position'];
    int bufferedPositionMs = positions['bufferedPosition'];
    return [
      Duration(milliseconds: positionMs),
      Duration(milliseconds: bufferedPositionMs)
    ];
  }

  Future<void> _packUp() async {
    await _channel.invokeMethod("packUp");
  }

  Future<void> _restore() async {
    value = VdoPlayerValue(isLoading: true);
    Map<String, Object> loadArgs = { 'restore' : true };
    Map loadResultMap = await _channel.invokeMethod("load", loadArgs);

    bool didRestore = loadResultMap['didLoad'];
    if (!didRestore) {
      print("Restored nothing");
    } else {
      print("restore success");
      Map mediaInfoMap = loadResultMap['mediaInfo'];

      MediaInfo mediaInfo = MediaInfo(
          mediaId: mediaInfoMap['mediaId'],
          type: MediaInfoType.streaming, // TODO fix
          title: mediaInfoMap['title'],
          description: mediaInfoMap['description'],
          duration: Duration(milliseconds: mediaInfoMap['duration']));

      value = value.copyWith(
          isLoading: false,
          mediaInfo: mediaInfo,
          duration: mediaInfo.duration,
          isPlaying: loadResultMap['playWhenReady']);

      if (value.isPlaying) {
        _startPositionTimer();
      }
    }
  }

  void _handleNativeEvent(nativeEvent) {
    if (nativeEvent is Map) {
      Map<dynamic, dynamic> event = nativeEvent;
      switch (event["name"]) {
        case "onPlayerStateChanged":
          final String playbackState = event["playbackState"];
          final bool ended = playbackState == 'ended';
          if (ended) {
            _stopPositionTimer();
          }
          // Set isEnded value here instead of in onMediaEnded, because it updates
          // if the playback is resumed after end
          value = value.copyWith(
            isBuffering: playbackState == 'buffering',
            isEnded: ended,
            position: ended ? value.duration : value.position
          );
          break;
        case "onMediaEnded":
          break;
        default:
          break;
      }
    }
  }

  void _handleNativeErrorEvent(nativeError) {
    if (nativeError is PlatformException) {
      PlatformException error = nativeError;
      print('native player error event: ' +
          error.code + ", " + error.message);
      _stopPositionTimer();
      value = value.copyWith(
          isLoading: false,
          isBuffering: false,
          vdoError: VdoError(int.parse(error.code), error.message, null)
      );
    }
  }
}
