import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:vdocipher_flutter/vdocipher_flutter.dart';

/// VdoPlayer controls created with [VdoPlayerController] to control playback and returns [onFullscreenChange] callback function
class VdoControllerView extends StatefulWidget {
  final VdoPlayerController controller;
  final Function(bool) onFullscreenChange;

  VdoControllerView({Key key, @required this.controller,
    this.onFullscreenChange}) : super(key: key);

  @override
  State<StatefulWidget> createState() => _VdoControllerViewState();
}

class _VdoControllerViewState extends State<VdoControllerView> {
  final Duration _skipInterval = Duration(milliseconds: 10000);

  VdoPlayerValue _playerValue;
  Size _windowSize;
  bool _isVisibleControls = true;
  double _iconSize = 40.0;
  bool _isFullScreenMode = false;

  @override
  void initState() {
    super.initState();
    _loadVideoInfo();
    widget.controller.addListener(_updatePlayerState);
    _updatePlayerState();
  }

  @override
  void dispose() {
    widget.controller.removeListener(_updatePlayerState);

    _portraitScreen();

    super.dispose();
  }

  _updatePlayerState() {
    setState(() {
      _playerValue = widget.controller.value;
    });
  }

  _loadVideoInfo() async {
    if (widget.controller == null) {
      return;
    }
  }

  Timer _hideTimer;

  @override
  Widget build(BuildContext context) {
    _windowSize = MediaQuery.of(context).size;
    bool showProgressIndicator =
        _playerValue.isLoading || _playerValue.isBuffering;
    bool showError = _playerValue.vdoError != null;
    bool showOverlayControls = _isVisibleControls && !showError;
    bool showBackgroundTint = showError || _isVisibleControls;

    return MouseRegion(
        onHover: (_) {
          _cancelAndRestartTimer();
        },
        child: GestureDetector(
            onTap: () => _cancelAndRestartTimer(),
            child: SizedBox(
              height: double.infinity,
              width: double.infinity,
              child: Stack(
                    alignment: AlignmentDirectional.bottomEnd,
                    fit: StackFit.expand,
                    children: [
                      // Background tint when controls are visible
                      Positioned(
                          child: !showBackgroundTint
                              ? SizedBox.shrink()
                              : Container(
                                  width: _windowSize.width,
                                  color: Colors.black.withOpacity(0.25))),

                      // Progress indicator
                      Positioned.fill(
                          child: !showProgressIndicator
                              ? SizedBox.shrink()
                              : Center(
                                  child:
                                      CircularProgressIndicator(strokeWidth: 2))),

                      // Error view
                      Positioned.fill(
                          child: !showError
                              ? SizedBox.shrink()
                              : Center(child: _errorInfo())),

                      // Video Overlay controllers
                      showOverlayControls
                          ? _vdoOverlayController()
                          : SizedBox.shrink(),

                      // Video seek bar
                      Positioned(
                        child: _isVisibleControls
                            ? _vdoProgress()
                            : SizedBox.shrink(),
                        bottom: 45.0,
                        left: 0,
                        right: 0,
                      ),

                      // Video bottom controllers
                      Positioned(
                          child: _isVisibleControls
                              ? _bottomControlPanel()
                              : SizedBox.shrink(),
                          bottom: 2.0,
                          left: 0,
                          right: 0),
                    ]
              ),
            )
        )
    );
  }

  void _cancelAndRestartTimer() {
    _hideTimer?.cancel();

    if (!mounted) {
      return;
    }

    // Visible controls on tap
    setState(() {
      _isVisibleControls = true;
    });

    // Hide controls after 3 seconds
    _hideTimer = Timer(const Duration(seconds: 3), () {
      // if video not playing then don't hide controls
      if (!widget.controller.isPlaying) {
        return;
      }

      if (!mounted) {
        return;
      }

      setState(() {
        _isVisibleControls = false;
      });
    });
  }

  /// Video Overlay controllers
  _vdoOverlayController() {
    return Center(
      child: Row(
        mainAxisAlignment: MainAxisAlignment.spaceAround,
        children: [
          InkWell(
              child: Icon(
                Icons.fast_rewind,
                color: Colors.white,
                size: _iconSize,
              ),
              onTap: () => _seekTo(-_skipInterval)
          ),
          InkWell(
              child: _playPauseRewind(),
              onTap: () => _playPauseRewindToggle()
          ),
          InkWell(
              child: Icon(
                Icons.fast_forward,
                color: Colors.white,
                size: _iconSize,
              ),
              onTap: () => _seekTo(_skipInterval)
          ),
        ],
      ),
    );
  }

  _errorInfo() {
    return Column(mainAxisAlignment: MainAxisAlignment.center, children: [
      Icon(Icons.error, size: _iconSize, color: Colors.white),
      Text('Error code: ' + _playerValue.vdoError.code.toString(),
          style: TextStyle(fontSize: 16, color: Colors.white))
    ]);
  }

  _playPauseRewind() {
    IconData icon = _playerValue.isEnded
        ? Icons.replay
        : _playerValue.isPlaying
            ? Icons.pause
            : Icons.play_arrow;
    return Icon(
      icon,
      color: Colors.white,
      size: _iconSize,
    );
  }

  /// Video Progress seekbar
  _vdoProgress() {
    return Container(
        height: 5,
        margin: EdgeInsets.only(left: 0.5, right: 1.0),
        child: SliderTheme(
            data: SliderTheme.of(context).copyWith(
              thumbShape: RoundSliderThumbShape(enabledThumbRadius: 4.5),
              overlayShape: RoundSliderOverlayShape(overlayRadius: 4.0),
              // trackHeight: 2.0
            ),
            child: Slider(
              value: _playerValue.position.inSeconds.toDouble(),
              min: 0,
              max: _playerValue.duration.inSeconds.toDouble(),
              activeColor: Colors.redAccent,
              inactiveColor: Colors.grey,
              onChanged: (val) async {
                print(val.round());
                await widget.controller.seek(Duration(seconds: val.round()));
              },
            )));
  }

  /// Video bottom controllers
  _bottomControlPanel() {
    return Row(mainAxisAlignment: MainAxisAlignment.spaceBetween, children: [
      Row(mainAxisAlignment: MainAxisAlignment.spaceEvenly, children: [
        Container(
            margin: EdgeInsets.only(left: 20.0, right: 20.0),
            child: Text(
                '${vdoTimeFormatter(_playerValue.position.inSeconds)} / ${vdoTimeFormatter(_playerValue.duration.inSeconds)}',
                style: TextStyle(fontSize: 20, color: Colors.white)))
      ]),
      Row(mainAxisAlignment: MainAxisAlignment.spaceEvenly, children: [
        _navItem(
            onClick: () => _toggleFullScreenMode(),
            iconData:
                _isFullScreenMode ? Icons.fullscreen_exit : Icons.fullscreen),
      ]),
    ]);
  }

  /// On Play/Pause button click handler for both overlay and bottom controller
  _playPauseRewindToggle() {
    if (widget.controller == null) {
      return;
    }

    if (_playerValue.isEnded) {
      widget.controller.seek(Duration.zero);
      widget.controller.play();
    } else if (_playerValue.isPlaying) {
      widget.controller.pause();
    } else {
      widget.controller.play();
    }
  }

  _seekTo(Duration interval) async {
    Duration currentPosition = await widget.controller.getPosition();
    widget.controller.seek(currentPosition + interval);
  }

  Container _navItem({Function onClick, IconData iconData}) {
    return Container(
        margin: EdgeInsets.only(left: 5.0, right: 5.0),
        child: InkWell(
          hoverColor: Colors.transparent,
          highlightColor: Colors.transparent,
          splashColor: Colors.transparent,
          onTap: onClick,
          child: Icon(
            iconData,
            size: _iconSize,
            color: Colors.white,
          ),
        ));
  }

  _toggleFullScreenMode() {
    setState(() {
      _isFullScreenMode = !_isFullScreenMode;
    });

    if (widget.onFullscreenChange != null) {
      widget.onFullscreenChange(_isFullScreenMode);
    }

    if (_isFullScreenMode) {
      SystemChrome.setEnabledSystemUIOverlays([]);
      SystemChrome.setPreferredOrientations(
          [DeviceOrientation.landscapeLeft, DeviceOrientation.landscapeRight]);
    } else {
      _portraitScreen();
    }
  }

  void _portraitScreen() {
    SystemChrome.setEnabledSystemUIOverlays(SystemUiOverlay.values);
    SystemChrome.setPreferredOrientations([DeviceOrientation.portraitUp]);
  }
}
