import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';
import 'package:flutter/material.dart';
import '../controllers/controllers.dart';

/// Widget for displaying video from your VdoCipher dashboard.
///
/// On being created, this widget creates a native platform video player and
/// provides a [VdoPlayerController] to control playback. The
/// [VdoPlayerController] reference is returned in the [onPlayerCreated] function
/// provided as an argument in the constructor.
class VdoPlatformView extends StatefulWidget {
  final Function(VdoPlayerController controller) onPlayerCreated;

  const VdoPlatformView({Key key, this.onPlayerCreated}) : super(key: key);

  @override
  State<StatefulWidget> createState() => _VdoPlatformViewState();
}

class _VdoPlatformViewState extends State<VdoPlatformView> {
  final String viewType = 'com.vdocipher.flutter/vdoplayerview';

  MethodChannel _channel;

  @override
  Widget build(BuildContext context) {
    return Container(
      width: double.infinity,
      height: double.infinity,
      child: _getPlatformView()
    );
  }

  _getPlatformView() {
    return PlatformViewLink(
      viewType: viewType,
      surfaceFactory:
          (BuildContext context, PlatformViewController controller) {
        return AndroidViewSurface(
          controller: controller,
          gestureRecognizers: const <Factory<OneSequenceGestureRecognizer>>{},
          hitTestBehavior: PlatformViewHitTestBehavior.opaque,
        );
      },
      onCreatePlatformView: (PlatformViewCreationParams params) {
        return PlatformViewsService.initSurfaceAndroidView(
          id: params.id,
          viewType: viewType,
          layoutDirection: TextDirection.ltr,
          creationParams: {},
          creationParamsCodec: StandardMessageCodec(),
        )
          ..addOnPlatformViewCreatedListener(
              _onPlatformViewCreated(params.onPlatformViewCreated))
          ..create();
      },
    );
  }

  _onPlatformViewCreated(Function platformViewCreatedCallback) {
    return ((int id) {
      print('vdoplayer platform view created, id = $id');
      _channel = MethodChannel('com.vdocipher.flutter.vdoplayerview_$id');
      _channel
          .setMethodCallHandler((call) => _handlePlatformMethodCall(call, id));
      platformViewCreatedCallback(id);
    });
  }

  Future<dynamic> _handlePlatformMethodCall(MethodCall call, int id) async {
    switch (call.method) {
      case 'onPlatformPlayerCreated':
        print('VdoPlayer: onPlatformPlayerCreated id $id');

        setState(() {
          VdoPlayerController vdoPlayerController = new VdoPlayerController(
              _channel,
              EventChannel('com.vdocipher.flutter.vdoplayerview.events_$id')
          );
          vdoPlayerController.init();
          widget.onPlayerCreated(vdoPlayerController);
        });
        break;
      default:
        throw MissingPluginException();
    }
  }
}
